/*
 * COPYRIGHT (c) International Business Machines Corp. 2001-2017
 *
 * This program is provided under the terms of the Common Public License,
 * version 1.0 (CPL-1.0). Any use, reproduction or distribution for this
 * software constitutes recipient's acceptance of CPL-1.0 terms which can be
 * found in the file LICENSE file or at
 * https://opensource.org/licenses/cpl1.0.php
 */

#ifndef OPENCRYPTOKI_PKCS11_H
#define OPENCRYPTOKI_PKCS11_H

#include <opencryptoki/pkcs11types.h>
#include <opencryptoki/apiclient.h>

#endif
